from abc import ABC, abstractmethod
from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
import pathlib
from typing import BinaryIO


class AzureBlobManager(ABC):

    def __init__(self, storage_account_name: str) -> None:
        self.account_url = f"https://{storage_account_name}.blob.core.windows.net"
        self.storage_account_name = storage_account_name
        self._authenticate_()
        self.blob_service_client: BlobServiceClient = None
        self.name: str = ''

    def __del__(self):
        if self.blob_service_client:
            self.blob_service_client.close()

    @abstractmethod
    def _authenticate_(self) -> None:
        pass

    @abstractmethod
    def _generate_blob_service_client_(self) -> None:
        """Generates a class property client for interacting with the whole blob service
        """
        pass

    @abstractmethod
    def _generate_container_client_(self, container_name: str) -> ContainerClient:
        """Generates a client for interacting with a specific container

        Args:
            container_name (str): The name of the container to interact with

        Returns:
            ContainerClient
        """
        pass

    def _generate_blob_client_(self, container_name: str, blob: str) -> BlobClient:
        """Generates and returns a client for interacting with a specific blob resource

        Args:
            container_name (str): The container where the resource lives

        Returns:
            BlobClient: The client to interact with the resource
        """
        if not self.blob_service_client:
            self._generate_blob_service_client_()

        return self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob
            )

    def upload_blob_from_file(
            self,
            file_path: str,
            container_name: str,
            **kwargs
            ) -> dict:
        """Uploads a system file as a blob file to a given container

        Args:
            file_path (str): The path where the file lives
            container_name (str): The container where the file will live

            :Keyword Arguments:
            **virtual_folder (str, optional): The name of the virtual folder where the blob will be uploaded
        """

        # file logic
        file = pathlib.Path(file_path)
        if not file.is_file():
            raise ValueError('The provided path was not found')

        # kwargs logic
        if 'virtual_folder' in kwargs:
            virtual_folder = kwargs.get('virtual_folder')
            if not virtual_folder:
                raise RuntimeError('Could not obtain the virutal folder name')
            blob_path = f'{virtual_folder}/{file.name}'

        blob_client = self._generate_blob_client_(container_name=container_name, blob=blob_path)

        # reading file
        with open(file_path, 'rb') as data:
            res = blob_client.upload_blob(data)

        return res

    def upload_blob_stream(
            self, 
            container_name: str,
            blob_name: str,
            data: BinaryIO,
            **kwargs) -> dict:
        """Uploads a binary file from an api service as a blob file to a given container\
        (TESTED ONLY ON FASTAPI)

        Args:
            container_name (str): The container where the blob will be uploaded to
            blob_name (str): The name of the blob file to be uploaded with
            data (BinaryIO): The binary data

            :Keyword Arguments:
            **virtual_folder (str, optional): The name of the virtual folder where the blob will be uploaded

        Returns:
            dict: _description_
        """

        # kwargs logic
        if 'virtual_folder' in kwargs:
            virtual_folder = kwargs.get('virtual_folder')
            if not virtual_folder:
                raise RuntimeError('Could not obtain the virutal folder name')
            blob_path = f'{virtual_folder}/{blob_name}'

        blob_client = self._generate_blob_client_(
            container_name=container_name,
            blob=blob_path)
        blob_client.upload_blob(data)

    def get_container_blob_names(self, container_name: str) -> list:
        """Retrives the names of the blob files in a given container

        Args:
            container_name (str): The container in which to look for

        Returns:
            list[str]: List of the blob names
        """
        container_client = self._generate_container_client_(
            container_name=container_name)
        blob_list = container_client.list_blob_names()
        return list(blob_list)

    def download_blob(self, container: str, blob_name: str) -> bytes:
        """Downloads bytes for a given blob resource

        Args:
            container (str): The container where the blob resource lives
            blob_name (str): The name of the blob resource, it should include the virtual folder if exists

        Returns:
            bytes: The content of the file
        """
        container_client = self._generate_container_client_(container_name=container)
        
        content: bytes = container_client.download_blob(blob=blob_name).readall()

        return content


import pytest
from .. import src
import os
import pymssql
from sqlalchemy.pool import NullPool
import mariadb


@pytest.fixture
def mssql_db():
    db = src.MsSQLDatabase(
            server=os.getenv('DB_mssql_SERVER'),
            user=os.getenv('DB_mssql_USER'),
            password=os.getenv('DB_mssql_PASSWORD'),
            database=os.getenv('DB_mssql_DATABASE'),
            pool_size=NullPool)
    yield db


@pytest.fixture
def mssql_db_error():
    db_error = src.MsSQLDatabase(
            server=os.getenv('DB_mssql_SERVER'),
            port=os.getenv('DB_mssql_PORT'),
            user=os.getenv('DB_mssql_USER_ERROR'),
            password=os.getenv('DB_mssql_PASSWORD_ERROR'),
            database=os.getenv('DB_mssql_DATABASE'))
    yield db_error


@pytest.fixture
def maria_db():
    db = src.MariaDatabase(
            server=os.getenv('DB_mariadb_SERVER'),
            user=os.getenv('DB_mariadb_USER'),
            password=os.getenv('DB_mariadb_PASSWORD'),
            database=os.getenv('DB_mariadb_DATABASE'),
            pool_size=NullPool)
    yield db


@pytest.fixture
def maria_db_error():
    db_error = src.MariaDatabase(
            server=os.getenv('DB_mariadb_SERVER'),
            user=os.getenv('DB_mariadb_USER_ERROR'),
            password=os.getenv('DB_mariadb_PASSWORD_ERROR'),
            database=os.getenv('DB_mariadb_DATABASE'))
    yield db_error


@pytest.mark.skip('Not implemented')
def test_mssql_db_error(mssql_db_error):
    with pytest.raises(pymssql._pymssql.OperationalError):
        mssql_db_error.connected


def test_maria_db_error(maria_db_error):
    with pytest.raises(mariadb.Error):
        with maria_db_error as conn:
            conn.connected


@pytest.mark.skip('Not implemented')
def test_mssql_db_success(mssql_db):
    assert mssql_db.connected


def test_maria_db_success(maria_db):
    with maria_db as conn:
        assert conn.connected


def test_maria_db_select(maria_db):
    with maria_db as conn:
        query = "SELECT * FROM tests.test WHERE id = ?"
        vals = (1,)
        res = conn.select(query=query, vals=vals)
        assert res == [{'id': 1, 'test_name': 'test1'}]


def test_maria_db_insert(maria_db):
    with maria_db as conn:
        query = """INSERT INTO tests.test(test_name) VALUES(?)"""
        data = ('test',)
        res = conn.insert(query=query, data=data)
        assert res == 1



class ConnectionException(Exception):
    """Raised when trying to access a connection that has not been declared
    """
    def __init__(self) -> None:
        super().__init__("The connection has not been properly set up")


class CursorException(Exception):
    """Raised when trying to access a cursos that has not been declared
    """
    def __init__(self) -> None:
        super().__init__("The Cursor has not been properly set up")


class SelectionException(Exception):
    """Raised when a SELECT query does not starts with 'SELECT'
    """
    def __init__(self) -> None:
        super().__init__("The query does not have the proper format. It should start with a 'SELECT' word")


class GroupException(Exception):
    """Raised when a group column is not in the selected columns
    """
    def __init__(self) -> None:
        super().__init__("Every Group column should be in the ")


class PoolSizeMinException(Exception):
    """Raised when a poolsize smaller than 1 is provided
    """
    def __init__(self) -> None:
        super().__init__("The minium size of a pool is 1")


class MaxOverflowSizeException(Exception):
    """Raised when the max_overflow is smaller than the poolsize
    """
    def __init__(self) -> None:
        super().__init__("The max overflow cannot be smaller than the requested pool")


class DataNotWritten(Exception):
    """Raised when the cursor does not return a rowcount > 0
    """
    def __init__(self, error) -> None:
        super().__init__(f"The data was not correcty inserted into the database. {error}")



from .baseclass import AzureBlobManager
import os
from azure.storage.blob import BlobServiceClient, ContainerClient

class AzureBlobManagementConnectionString(AzureBlobManager):
    
    def __init__(self, storage_account_name: str) -> None:
        super().__init__(storage_account_name)
        self.name = 'connectionString'

    def _authenticate_(self) -> None:
        self.connect_str = os.getenv('AZURE_STORAGE_CONNECTION_STRING')

    def _generate_blob_service_client_(self) -> None:
        self.blob_service_client = BlobServiceClient.from_connection_string(self.connect_str)

    def _generate_container_client_(self, container_name: str) -> ContainerClient:
        return ContainerClient.from_connection_string(conn_str=self.connect_str, container_name=container_name)